@@ -18,19 +18,19 @@
 */
 
 #include "includes.h"
-#include "torture/torture.h"
 #include "system/filesys.h"
 #include "system/dir.h"
-#include "../tdb/include/tdb.h"
+#include "../lib/tdb_compat/tdb_compat.h"
 #include "../lib/util/util_tdb.h"
 #include "libcli/libcli.h"
 #include "torture/util.h"
+#include "torture/basic/proto.h"
 
 static TDB_CONTEXT *tdb;
 
 #define NAME_LENGTH 20
 
-static uint_t total, collisions, failures;
+static unsigned int total, collisions, failures;
 
 static bool test_one(struct torture_context *tctx ,struct smbcli_state *cli, 
 		     const char *name)
@@ -115,8 +115,8 @@ static bool test_one(struct torture_context *tctx ,struct smbcli_state *cli,
 static char *gen_name(TALLOC_CTX *mem_ctx)
 {
 	const char *chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz._-$~...";
-	uint_t max_idx = strlen(chars);
-	uint_t len;
+	unsigned int max_idx = strlen(chars);
+	unsigned int len;
 	int i;
 	char *p;
 	char *name;
@@ -167,7 +167,7 @@ bool torture_mangle(struct torture_context *torture,
 	int i;
 
 	/* we will use an internal tdb to store the names we have used */
-	tdb = tdb_open(NULL, 100000, TDB_INTERNAL, 0, 0);
+	tdb = tdb_open_compat(NULL, 100000, TDB_INTERNAL, 0, 0, NULL, NULL);
 	if (!tdb) {
 		printf("ERROR: Failed to open tdb\n");
 		return false;
